@@ -1,78 +1,38 @@
 /* SPDX-License-Identifier: (Apache-2.0 OR MIT OR Zlib) */
 /* Copyright © 2021 Violet Leonard */
 
-use std::cell::Cell;
-use std::ops::{Deref, DerefMut};
+use std::{
+    fmt,
+    ops::{Deref, DerefMut},
+};
 
-use super::{WatchContext, WatchSet};
-
-/// This provides the basic functionality behind watched values. You can use
-/// it to provide functionality using the watch system for cases where
-/// [Watched](struct.Watched.html) and
-/// [WatchedEvent](struct.WatchedEvent.html) are not appropriate.
-#[derive(Default)]
-pub struct WatchedMeta {
-    watchers: WatchSet,
-}
-
-impl WatchedMeta {
-    /// Create a new WatchedMeta instance
-    pub fn new() -> Self {
-        WatchedMeta {
-            watchers: WatchSet::new(),
-        }
-    }
-
-    /// When run in a function designed to watch a value, will bind so that
-    /// function will be re-run when this is triggered.
-    pub fn watched(&self) {
-        WatchContext::try_get_current(|ctx| {
-            if let Some(watch) = ctx.current_watch() {
-                self.watchers.add(watch);
-            }
-        });
-    }
-
-    /// Mark this value as having changed, so that watching functions will
-    /// be marked as needing to be updated.
-    /// # Panics
-    /// This function will panic if called outside of WatchContext::with
-    pub fn trigger(&self) {
-        WatchContext::expect_current(
-            |ctx| {
-                ctx.add_to_next(&self.watchers);
-            },
-            "WatchedMeta.trigger() called outside of WatchContext",
-        );
-    }
-}
+use crate::{DefaultOwner, WatchedCellCore, WatchedCore};
 
 /// This represents some value which will be interesting to watch. Watcher
 /// functions that reference this value will be re-run when this value
 /// changes.
+#[derive(Default)]
 pub struct Watched<T: ?Sized> {
-    meta: WatchedMeta,
-    value: T,
+    inner: WatchedCore<'static, T, DefaultOwner>,
 }
 
 impl<T> Watched<T> {
     /// Create a new watched value.
     pub fn new(value: T) -> Self {
         Watched {
-            value,
-            meta: WatchedMeta::new(),
+            inner: WatchedCore::new(value),
         }
     }
 
     /// Consumes the `Watched`, returning the wrapped value
-    pub fn into_inner(self) -> T {
-        self.value
+    pub fn into_inner(this: Self) -> T {
+        this.inner.into_inner()
     }
 
     /// Replaces the wrapped value with a new one, returning the old value,
     /// without deinitializing either one.
-    pub fn replace(&mut self, value: T) -> T {
-        std::mem::replace(&mut *self, value)
+    pub fn replace(this: &mut Self, value: T) -> T {
+        std::mem::replace(&mut *this, value)
     }
 }
 
@@ -80,14 +40,14 @@ impl<T: ?Sized> Watched<T> {
     /// Get a referenced to the wrapped value, without binding the current
     /// watch closure.
     pub fn get_unwatched(this: &Self) -> &T {
-        &this.value
+        this.inner.get_unwatched()
     }
 }
 
 impl<T: Default> Watched<T> {
     /// Takes the wrapped value, leaving `Default::default()` in its place.
-    pub fn take(&mut self) -> T {
-        std::mem::take(&mut *self)
+    pub fn take(this: &mut Self) -> T {
+        std::mem::take(&mut *this)
     }
 }
 
@@ -101,44 +61,44 @@ impl<T: PartialEq> Watched<T> {
     /// sync. This would normally cause an infinite loop as each update of
     /// one would cause the other one to re-evaluate. However using set_if_neq
     /// allows it to detect that the value is the same and stop propogating.
+    ///
     /// ```rust
-    /// # use drying_paint::*;
+    ///# use std::{rc::Rc, cell::RefCell};
+    ///# use drying_paint::{Watcher, Watched, WatcherInit, WatchContext};
     /// #[derive(Default)]
     /// struct KeepBalanced {
     ///     left: Watched<i32>,
     ///     right: Watched<i32>,
     /// }
-    /// impl WatcherInit for KeepBalanced {
-    ///     fn init(watcher: &mut WatcherMeta<Self>) {
-    ///         watcher.watch(|root| {
+    ///
+    /// impl Watcher<'static> for KeepBalanced {
+    ///     fn init(mut init: impl WatcherInit<'static, Self>) {
+    ///         init.watch(|root| {
     ///             Watched::set_if_neq(&mut root.left, *root.right);
     ///         });
-    ///         watcher.watch(|root| {
+    ///         init.watch(|root| {
     ///             Watched::set_if_neq(&mut root.right, *root.left);
     ///         });
     ///     }
     /// }
-    /// fn main() {
-    ///     let mut ctx = WatchContext::new();
-    ///     ctx.set_frame_limit(Some(100));
-    ///     ctx.with(|| {
-    ///         let obj = WatchContext::allow_watcher_access((), |()| {
-    ///             let mut obj = Watcher::<KeepBalanced>::new();
-    ///             *obj.data_mut().left = 68;
-    ///             obj
-    ///         });
-    ///         WatchContext::update_current();
-    ///         WatchContext::allow_watcher_access(obj, |obj| {
-    ///             assert_eq!(*obj.data().right, 68);
-    ///         });
-    ///     });
-    /// }
+    ///
+    /// let keep_balanced = Rc::new(RefCell::new(KeepBalanced {
+    ///     left: Watched::new(7),
+    ///     right: Watched::new(7),
+    /// }));
+    /// let weak = Rc::downgrade(&keep_balanced);
+    /// let mut ctx = WatchContext::new();
+    /// ctx.set_frame_limit(Some(10));
+    /// ctx.add_watcher(&weak);
+    /// *keep_balanced.borrow_mut().left = 3;
+    /// ctx.update();
+    /// assert_eq!(keep_balanced.borrow().right, 3);
+    /// *keep_balanced.borrow_mut().right = 21;
+    /// ctx.update();
+    /// assert_eq!(keep_balanced.borrow().left, 21);
+    /// ```
     pub fn set_if_neq(wrapper: &mut Watched<T>, value: T) {
-        if wrapper.value != value {
-            wrapper.value = value;
-            wrapper.meta.trigger();
-            wrapper.meta.watched();
-        }
+        wrapper.inner.set_if_neq_auto(value);
     }
 }
 
@@ -146,87 +106,39 @@ impl<T: ?Sized> Deref for Watched<T> {
     type Target = T;
 
     fn deref(&self) -> &T {
-        self.meta.watched();
-        &self.value
+        self.inner.get_auto()
     }
 }
 
 impl<T: ?Sized> DerefMut for Watched<T> {
     fn deref_mut(&mut self) -> &mut T {
-        self.meta.trigger();
-        self.meta.watched();
-        &mut self.value
+        self.inner.get_mut_auto()
     }
 }
 
-impl<T: Default> Default for Watched<T> {
-    fn default() -> Self {
-        Watched::new(Default::default())
-    }
-}
-
-use std::fmt;
-
 impl<T: fmt::Debug + ?Sized> fmt::Debug for Watched<T> {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        fmt::Debug::fmt(&**self, f)
-    }
-}
-
-impl<T: fmt::Display + ?Sized> fmt::Display for Watched<T> {
-    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        fmt::Display::fmt(&**self, f)
-    }
-}
-
-impl<T: Clone> Clone for Watched<T> {
-    fn clone(&self) -> Watched<T> {
-        Watched {
-            value: Clone::clone(&self.value),
-            meta: WatchedMeta::new(),
-        }
-    }
-}
-
-#[cfg(feature = "serde")]
-impl<T: serde::Serialize + ?Sized> serde::Serialize for Watched<T> {
-    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
-    where
-        S: serde::Serializer,
-    {
-        T::serialize(&self.value, serializer)
-    }
-}
-
-#[cfg(feature = "serde")]
-impl<'de, T> serde::Deserialize<'de> for Watched<T>
-where
-    T: serde::Deserialize<'de> + ?Sized,
-{
-    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
-    where
-        D: serde::Deserializer<'de>,
-    {
-        T::deserialize(deserializer).map(Self::new)
+        fmt::Debug::fmt(self.inner.get_auto(), f)
     }
 }
 
 mod watched_ops {
-    use core::cmp::Ordering;
-    use core::ops::*;
+    use std::cmp::Ordering;
+    use std::ops::*;
 
     use super::Watched;
 
     macro_rules! watched_unop {
         (impl $imp:ident, $method:ident) => {
+            /*
             impl<T: $imp> $imp for Watched<T> {
                 type Output = <T as $imp>::Output;
 
                 fn $method(self) -> <T as $imp>::Output {
-                    self.meta.watched();
-                    $imp::$method(self.value)
+                    $imp::$method(self.inner.get_auto())
                 }
             }
+            */
 
             impl<'a, T: ?Sized> $imp for &'a Watched<T>
             where
@@ -235,8 +147,7 @@ mod watched_ops {
                 type Output = <&'a T as $imp>::Output;
 
                 fn $method(self) -> <&'a T as $imp>::Output {
-                    self.meta.watched();
-                    $imp::$method(&self.value)
+                    $imp::$method(self.inner.get_auto())
                 }
             }
         };
@@ -244,6 +155,7 @@ mod watched_ops {
 
     macro_rules! watched_binop {
         (impl $imp:ident, $method:ident) => {
+            /*
             impl<T, U> $imp<U> for Watched<T>
             where
                 T: $imp<U>,
@@ -251,10 +163,10 @@ mod watched_ops {
                 type Output = <T as $imp<U>>::Output;
 
                 fn $method(self, other: U) -> <T as $imp<U>>::Output {
-                    self.meta.watched();
-                    $imp::$method(self.value, other)
+                    $imp::$method(self.inner.get_auto(), other)
                 }
             }
+            */
 
             impl<'a, T, U> $imp<U> for &'a Watched<T>
             where
@@ -264,8 +176,7 @@ mod watched_ops {
                 type Output = <&'a T as $imp<U>>::Output;
 
                 fn $method(self, other: U) -> <&'a T as $imp<U>>::Output {
-                    self.meta.watched();
-                    $imp::$method(&self.value, other)
+                    $imp::$method(self.inner.get_auto(), other)
                 }
             }
         };
@@ -278,10 +189,7 @@ mod watched_ops {
                 T: $imp<U> + ?Sized,
             {
                 fn $method(&mut self, rhs: U) {
-                    let res = $imp::$method(&mut self.value, rhs);
-                    self.meta.trigger();
-                    self.meta.watched();
-                    res
+                    $imp::$method(self.inner.get_mut_auto(), rhs);
                 }
             }
         };
@@ -318,14 +226,12 @@ mod watched_ops {
         U: ?Sized,
     {
         fn eq(&self, other: &U) -> bool {
-            self.meta.watched();
-            PartialEq::eq(&self.value, other)
+            PartialEq::eq(self.inner.get_auto(), other)
         }
 
         #[allow(clippy::partialeq_ne_impl)]
         fn ne(&self, other: &U) -> bool {
-            self.meta.watched();
-            PartialEq::ne(&self.value, other)
+            PartialEq::ne(self.inner.get_auto(), other)
         }
     }
 
@@ -335,33 +241,26 @@ mod watched_ops {
         U: ?Sized,
     {
         fn partial_cmp(&self, other: &U) -> Option<Ordering> {
-            self.meta.watched();
-            PartialOrd::partial_cmp(&self.value, other)
+            PartialOrd::partial_cmp(self.inner.get_auto(), other)
         }
         fn lt(&self, other: &U) -> bool {
-            self.meta.watched();
-            PartialOrd::lt(&self.value, other)
+            PartialOrd::lt(self.inner.get_auto(), other)
         }
         fn le(&self, other: &U) -> bool {
-            self.meta.watched();
-            PartialOrd::le(&self.value, other)
+            PartialOrd::le(self.inner.get_auto(), other)
         }
         fn ge(&self, other: &U) -> bool {
-            self.meta.watched();
-            PartialOrd::ge(&self.value, other)
+            PartialOrd::ge(self.inner.get_auto(), other)
         }
         fn gt(&self, other: &U) -> bool {
-            self.meta.watched();
-            PartialOrd::gt(&self.value, other)
+            PartialOrd::gt(self.inner.get_auto(), other)
         }
     }
 
     /*
     impl<T: Ord> Ord for Watched<T> {
         fn cmp(&self, other: &Watched<T>) -> Ordering {
-            self.meta.watched();
-            other.meta.watched();
-            Ord::cmp(&self.value, &other.value)
+            Ord::cmp(self.inner.get_auto(), other.inner.get_auto())
         }
     }
 
@@ -375,8 +274,7 @@ mod watched_ops {
 /// `RefCell<Watched<T>>`.
 #[derive(Default)]
 pub struct WatchedCell<T: ?Sized> {
-    meta: WatchedMeta,
-    value: Cell<T>,
+    inner: WatchedCellCore<'static, T, DefaultOwner>,
 }
 
 impl<T: ?Sized> WatchedCell<T> {
@@ -385,14 +283,12 @@ impl<T: ?Sized> WatchedCell<T> {
     /// This call borrows the WatchedCell mutably (at compile-time) which
     /// guarantees that we possess the only reference.
     pub fn get_mut(&mut self) -> &mut T {
-        self.meta.trigger();
-        self.meta.watched();
-        self.value.get_mut()
+        self.inner.get_mut_auto()
     }
 
     /// Treat this WatchedCell as watched, without fetching the actual value.
     pub fn watched(&self) {
-        self.meta.watched();
+        self.inner.watched_auto();
     }
 }
 
@@ -400,50 +296,43 @@ impl<T> WatchedCell<T> {
     /// Create a new WatchedCell
     pub fn new(value: T) -> Self {
         Self {
-            meta: WatchedMeta::new(),
-            value: Cell::new(value),
+            inner: WatchedCellCore::new(value),
         }
     }
 
     /// Sets the watched value
     pub fn set(&self, value: T) {
-        self.replace(value);
+        self.inner.set_auto(value);
     }
 
     /// Unwraps the WatchedCell, returning the contained value
     pub fn into_inner(self) -> T {
-        self.value.into_inner()
+        self.inner.into_inner()
     }
 
     /// Replaces the contained value and returns the previous value
     pub fn replace(&self, value: T) -> T {
-        self.meta.trigger();
-        self.meta.watched();
-        self.value.replace(value)
+        self.inner.replace_auto(value)
     }
 }
 
 impl<T: Copy> WatchedCell<T> {
     /// Returns a copy of the watched value
     pub fn get(&self) -> T {
-        self.meta.watched();
-        self.value.get()
+        self.inner.get_auto()
     }
 }
 
 impl<T: Default> WatchedCell<T> {
     /// Takes the watched value, leaving `Default::default()` in its place
     pub fn take(&self) -> T {
-        self.meta.trigger();
-        self.meta.watched();
-        self.value.take()
+        self.inner.take_auto()
     }
 }
 
 impl<T: Copy> Clone for WatchedCell<T> {
     fn clone(&self) -> Self {
-        self.meta.watched();
-        Self::new(self.value.get())
+        Self::new(self.get())
     }
 }
 
@@ -455,93 +344,65 @@ impl<T> From<T> for WatchedCell<T> {
 
 #[cfg(test)]
 mod tests {
+    use std::{cell::RefCell, rc::Rc};
+
     use crate::*;
 
     #[test]
-    fn test_watched_add() {
+    fn watched_add() {
         let left = Watched::new(587);
-        assert_eq!((&left) + 13, 600);
-        assert_eq!(left + 13, 600);
-    }
-
-    #[derive(Default)]
-    struct Inner {
-        value: Watched<u32>,
-    }
-
-    type Outer = Watcher<OuterData>;
-
-    #[derive(Default)]
-    struct OuterData {
-        value: u32,
-        inner: Inner,
-    }
-
-    impl WatcherInit for OuterData {
-        fn init(watcher: &mut WatcherMeta<Self>) {
-            watcher.watch(|root| {
-                root.value = *root.inner.value;
-            });
-        }
+        assert_eq!(&left + 13, 600);
     }
 
     #[test]
-    fn test_add_to_watched() {
-        let mut ctx = WatchContext::new();
-        ctx = ctx
-            .with(|| {
-                let outer = WatchContext::allow_watcher_access((), |()| {
-                    let mut outer = Outer::new();
-                    *outer.data_mut().inner.value = 587;
-                    outer
-                });
-                WatchContext::update_current();
-                let outer =
-                    WatchContext::allow_watcher_access(outer, |mut outer| {
-                        assert_eq!(outer.data().value, 587);
-                        outer.data_mut().inner.value += 13;
-                        outer
-                    });
-                WatchContext::update_current();
-                WatchContext::allow_watcher_access(outer, |outer| {
-                    assert_eq!(outer.data().value, 600);
-                });
-            })
-            .0;
-        std::mem::drop(ctx);
-    }
-
-    #[derive(Default)]
-    struct OuterXorData {
-        value: u32,
-        inner: Inner,
-    }
-
-    impl WatcherInit for OuterXorData {
-        fn init(watcher: &mut WatcherMeta<Self>) {
-            watcher.watch(|root| {
-                root.value = &root.inner.value ^ 0xffffffff;
-            });
+    fn add_to_watched() {
+        struct Content {
+            dest: u32,
+            source: Watched<u32>,
         }
+
+        impl Watcher<'static> for Content {
+            fn init(mut init: impl WatcherInit<'static, Self>) {
+                init.watch(|root| {
+                    root.dest = *root.source;
+                });
+            }
+        }
+        let content = Rc::new(RefCell::new(Content {
+            dest: 0,
+            source: Watched::new(587),
+        }));
+        let weak = Rc::downgrade(&content);
+        let mut ctx = WatchContext::new();
+        ctx.add_watcher(&weak);
+        assert_eq!(content.borrow().dest, 587);
+        content.borrow_mut().source += 13;
+        ctx.update();
+        assert_eq!(content.borrow().dest, 600);
     }
 
     #[test]
-    fn test_xor_watch() {
+    fn watched_xor() {
+        #[derive(Default)]
+        struct Content {
+            dest: u32,
+            source: Watched<u32>,
+        }
+
+        impl Watcher<'static> for Content {
+            fn init(mut init: impl WatcherInit<'static, Self>) {
+                init.watch(|root| {
+                    root.dest = &root.source ^ 0xffffffff;
+                });
+            }
+        }
+        let content = Rc::new(RefCell::new(Content::default()));
+        let weak = Rc::downgrade(&content);
         let mut ctx = WatchContext::new();
-        ctx = ctx
-            .with(|| {
-                let outer = WatchContext::allow_watcher_access((), |()| {
-                    let mut outer = Watcher::<OuterXorData>::new();
-                    *outer.data_mut().inner.value = 960294194;
-                    outer
-                });
-                WatchContext::update_current();
-                WatchContext::allow_watcher_access(outer, |outer| {
-                    assert_eq!(outer.data().value, 3334673101);
-                });
-            })
-            .0;
-        std::mem::drop(ctx);
+        ctx.add_watcher(&weak);
+        *content.borrow_mut().source = 960294194;
+        ctx.update();
+        assert_eq!(content.borrow().dest, 3334673101);
     }
 
     #[test]
